
#include "codegen.h"
#include "stringbuffer.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#define NEW(_STRUCT) (_STRUCT*)malloc(sizeof(_STRUCT))

/* *************************************************************** Strukturen */

typedef struct
{
	stringBuffer* content;   /**< Inhalt dieser Schicht */
	int stackSize,           /**< aktuelle Stackgröße */
	    stackSizeMax;        /**< maximale Stackgröße */
	const void* index;       /**< interne Adresse für den Methodenkopf */
} codegenLayer;

struct sCodegen
{
	stringBuffer *staticVariables; /**< Puffer für die globalen Variablen */
	codegenLayer local,            /**< methodenlokaler Code */
	             global,           /**< programmglobaler Code */
	             *curr;            /**< aktueller Gültigkeitsbereich */
	unsigned int label,            /**< aktuelle Labelnummer */
	             localVarCount,    /**< maximale Anzahl gleichzeitig benutzter
	                                    lokaler Variablen */
	             localVar;         /**< aktuelle Anzahl gleichzeitig benutzter
	                                    lokaler Variablen */
	char* className;               /**< Name der zu generierenden Klasse */
};

/* ******************************************************* private Funktionen */

/**@brief Gibt die Jasmin-Codes für die jeweiligen Rückgabetypen zurück.
 */
static char convertType(int type)
{
	switch (type)
	{
	case MINAKO_TYPE_INT:
		return 'I';
		
	case MINAKO_TYPE_FLOAT:
		return 'F';
		
	case MINAKO_TYPE_BOOL:
		return 'Z';
		
	case MINAKO_TYPE_VOID:
		return 'V';
		
	default:
		yyerror("Unbekannter Typ");
	}
	
	return 0;
}

/* *************************************************** exportierte Funktionen */

codegen* newCodegen(const char* className)
{
	codegen* gen = NEW(codegen);
	
	assert(className);
	
	/* Puffer anlegen */
	gen->staticVariables = newStringBuffer();
	gen->global.content = newStringBuffer();
	gen->local.content = newStringBuffer();
	
	/* globale Stackgrößen initialisieren */
	gen->global.stackSize = gen->global.stackSizeMax = 0;
	
	/* aktuellen Puffer auf den globalen setzen */
	gen->curr = &gen->global;
	
	gen->className = strdup(className);
	
	stringBufferAppend(gen->staticVariables, "; ------- Klassenvariablen ----------------------------------");
	
	stringBufferAppend(gen->global.content, "; ------- Klasseninitialisierer -----------------------------");
	stringBufferAppend(gen->global.content, ".method static <clinit>()V");
	gen->global.index = stringBufferGetCurrentAddress(gen->global.content);
	stringBufferAppend(gen->global.content, "\t.limit locals 0");
	
	return gen;
}

void deleteCodegen(codegen* gen)
{
	char tmp[32];
	assert(gen);
	
	/* rausschreiben aller generierten Daten */
	fprintf(yyout, "; ------- Header --------------------------------------------\n");
	fprintf(yyout, ".class  public synchronized %s\n", gen->className);
	fprintf(yyout, ".super  java/lang/Object\n");
	fprintf(yyout, "; -----------------------------------------------------------\n\n");
	
	stringBufferPrint(gen->staticVariables, yyout);
	fprintf(yyout, "; -----------------------------------------------------------\n\n");
	
	fprintf(yyout, "; ------- Constructor ---------------------------------------\n");
	fprintf(yyout, ".method public <init>()V\n");
	fprintf(yyout, "\t.limit stack 1\n");
	fprintf(yyout, "\t.limit locals 1\n");
	fprintf(yyout, "\taload_0\n");
	fprintf(yyout, "\tinvokenonvirtual java/lang/Object/<init>()V\n");
	fprintf(yyout, "\treturn\n");
	fprintf(yyout, ".end method\n");
	fprintf(yyout, "; -----------------------------------------------------------\n\n");
	
	stringBufferPrint(gen->local.content, yyout);
	
	sprintf(tmp, "\t.limit stack %d", gen->global.stackSizeMax);
	stringBufferInsert(gen->global.content, gen->global.index, tmp);
	stringBufferPrint(gen->global.content, yyout);
	fprintf(yyout, "\treturn\n");
	fprintf(yyout, ".end method\n");
	fprintf(yyout, "; -----------------------------------------------------------\n");
	
	/* aufräumen */
	deleteStringBuffer(gen->staticVariables);
	deleteStringBuffer(gen->global.content);
	deleteStringBuffer(gen->local.content);
	free(gen->className);
	free(gen);
}

void codegenAppendCommand(codegen* gen, const char* cmd, int stackDiff)
{
	char* tmp;
	assert(gen && cmd);
	
	/* berechnen der korrekten Stringlänge für den Puffer */
	tmp = (char*) malloc(sizeof(char) * strlen(cmd) + sizeof("\t"));
	sprintf(tmp, "\t%s", cmd);
	
	/* an den aktuellen Puffer anhängen */
	stringBufferAppend(gen->curr->content, tmp);
	
	free(tmp);
	
	/* Stackgröße berechnen */
	gen->curr->stackSize += stackDiff;
	
	if (gen->curr->stackSize > gen->curr->stackSizeMax)
		gen->curr->stackSizeMax = gen->curr->stackSize;
}

void codegenInsertCommand(codegen* gen, const void* address, const char* cmd,
                          int stackDiff)
{
	char* tmp;
	assert(gen && cmd);
	
	/* berechnen der korrekten Stringgröße */
	tmp = (char*) malloc(sizeof(char) * strlen(cmd) + sizeof("\t"));
	sprintf(tmp, "\t%s", cmd);
	
	/* in den aktuellen Puffer einfügen */
	stringBufferInsert(gen->curr->content, address, tmp);
	
	free(tmp);
	
	/* Stackgröße berechnen */
	gen->curr->stackSize += stackDiff;
	
	if (gen->curr->stackSize > gen->curr->stackSizeMax)
		gen->curr->stackSizeMax = gen->curr->stackSize;
}

void codegenAppendLabel(codegen* gen, unsigned int label)
{
	char tmp[32];
	assert(gen);
	
	sprintf(tmp, "Label%u:", label);
	stringBufferAppend(gen->curr->content, tmp);
}

void codegenAddVariable(codegen* gen, const char* name, minakoVariable* var)
{
	char* tmp;
	assert(gen && name);
	
	/* Testen, ob wir gerade in der globalen Schicht sind */
	if (gen->curr == &gen->global)
	{
		/* berechnen der korrekten Stringgröße */
		tmp = (char*) malloc(sizeof(".field static " /* Name */ " _") +
			                   sizeof(char)*strlen(name));
		
		sprintf(tmp, ".field static %s %c", name, convertType(var->type));
		stringBufferAppend(gen->staticVariables, tmp);
		
		free(tmp);
		
		/* die Dummyadresse signalisiert, dass es sich um eine globale Variable
		 * handelt */
		var->address = -1;
	}
	else
	{
		/* neue Jasmin-interne Adresse zuweisen */
		var->address = gen->localVar;
		
		/* Anzahl der lokalen Variablen erhöhen */
		++gen->localVar;
		
		if (gen->localVar > gen->localVarCount)
		  gen->localVarCount = gen->localVar;
	}
}

void codegenRemoveVariable(codegen* gen, minakoVariable* var)
{
	assert(gen && var);
	
	/* testen, ob wir eine lokale Variable abräumen wollen */
	/* (dieser Test ist im Prinzip überflüssig, da die globale Schicht erst
	 *  zuletzt freigegeben wird) */
	if (var->address >= 0)
		--gen->localVar;
}

void codegenAssignVariable(codegen* gen, const char* name, minakoVariable* var)
{
	char* tmp;
	assert(gen && var);
	
	/* teste, ob die Variable lokal ist */
	if (var->address >= 0)
	{
		tmp = (char*) malloc(sizeof(char) * 32);
		
		/* generiere die entsprechende Anweisung */
		switch (var->type)
		{
		case MINAKO_TYPE_BOOL:
		case MINAKO_TYPE_INT:
			sprintf(tmp, "istore %d", var->address);
			break;
			
		case MINAKO_TYPE_FLOAT:
			sprintf(tmp, "fstore %d", var->address);
		}
	}
	else
	{
		/* Bestimmen der Stringgröße */
		tmp = (char*) malloc(sizeof("putstatic " /* Klasse */ "/" /* Name */ " _") +
		                     sizeof(char) *(strlen(gen->className) + strlen(name)));
		
		/* Generieren der Anweisung */
		sprintf(tmp, "putstatic %s/%s %c", gen->className, name,
		                                   convertType(var->type));
	}
	
	codegenAppendCommand(gen, tmp, -1);
	free(tmp);
}

unsigned int codegenNextLabel(codegen* gen)
{
	assert(gen);
	return ++gen->label;
}

const void* codegenCurrentAddress(const codegen* gen)
{
	assert(gen);
	return stringBufferGetCurrentAddress(gen->curr->content);
}

void codegenEnterFunction(codegen* gen, const char* name,
                          const minakoFunction* func)
{
	assert(gen && func);
	
	/* aktuellen Gültigkeitsbereich auf methodenlokal setzen */
	gen->curr = &gen->local;
	
	/* Labelzähler neu initialisieren */
	gen->label = 0;
	
	/* lokale Stackgröße initialisieren */
	gen->local.stackSize = gen->local.stackSizeMax = 0;
	
	/* Test, ob wir die main-Funktion in der Hand haben */
	if (!strcmp(name, "main"))
	{
		/* Sonderbehandlung für diesen Fall */
		gen->localVarCount = gen->localVar = 1;
		
		stringBufferAppend(gen->local.content, "; ------- Methode ---- void main() --------------------------");
		stringBufferAppend(gen->local.content, ".method public static main([Ljava/lang/String;)V");
	}
	else
	{
		int i;
		/* berechnen der korrekten Stringgröße (etwas aufwendiger hier) */
		char *tmp =
			(char*) malloc(sizeof(".method public static " /* Name */
			                      "(" /* Parameter */ ")_") + 
		                 sizeof(char) * (strlen(name) + func->argCount)),
		     *ptr;
		
		/* lokale Variablenanzahl initialisieren */
		gen->localVarCount = gen->localVar = 0;
		
		stringBufferAppend(gen->local.content, "; ------- Methode -------------------------------------------");
		
		ptr = tmp + sprintf(tmp, ".method public static %s(", name);
		i = 0;
		while (i < func->argCount)
		{
			minakoVariable* param = (minakoVariable*) func->param[i]->data;
			
			*ptr++ = convertType(param->type);
			++i;
		}
		
		*ptr++ = ')';
		*ptr++ = convertType(func->type);
		*ptr = '\0';
		
		/* anhängen der Deklarationszeile */
		stringBufferAppend(gen->local.content, tmp);
		
		free(tmp);
	}
	
	gen->local.index = stringBufferGetCurrentAddress(gen->local.content);
}

void codegenLeaveFunction(codegen* gen)
{
	char tmp[32];
	assert(gen);
	
	/* Setzen des aktuellen Gültigkeitsbereiches auf global */
	gen->curr = &gen->global;
	
	sprintf(tmp, "\t.limit locals %d", gen->localVarCount);
	stringBufferInsert(gen->local.content, gen->local.index, tmp);
	
	sprintf(tmp, "\t.limit stack %d", gen->local.stackSizeMax);
	stringBufferInsert(gen->local.content, gen->local.index, tmp);
	
	stringBufferAppend(gen->local.content, "\treturn");
	stringBufferAppend(gen->local.content, ".end method");
	stringBufferAppend(gen->local.content, "; -----------------------------------------------------------\n");
}
